""" Copyright 2012 INRIA

    Author: Louis-Marie Traonouez <Louis-Marie.Traonouez@inria.fr>

    This software is governed by the CeCILL  license under French law and
    abiding by the rules of distribution of free software.  You can  use, 
    modify and/ or redistribute the software under the terms of the CeCILL
    license as circulated by CEA, CNRS and INRIA at the following URL
    "http://www.cecill.info". 
    
    As a counterpart to the access to the source code and  rights to copy,
    modify and redistribute granted by the license, users are provided only
    with a limited warranty  and the software's author,  the holder of the
    economic rights,  and the successive licensors  have only  limited
    liability. 
    
    In this respect, the user's attention is drawn to the risks associated
    with loading,  using,  modifying and/or developing or reproducing the
    software by the user in light of its specific status of free software,
    that may mean  that it is complicated to manipulate,  and  that  also
    therefore means  that it is reserved for developers  and  experienced
    professionals having in-depth computer knowledge. Users are therefore
    encouraged to load and test the software's suitability as regards their
    requirements in conditions enabling the security of their systems and/or 
    data to be ensured and,  more generally, to use and operate it in the 
    same conditions as regards security. 
    
    The fact that you are presently reading this means that you have had
    knowledge of the CeCILL license and that you accept its terms.
 """


import sys
import logging
import copy

import pynta
import pydbm
from pyparsing import Word, Combine, Optional, Forward, Group, Literal, Keyword, ZeroOrMore, OneOrMore, Suppress, alphas, alphanums, nums, stringEnd

logging.basicConfig()
logger = logging.getLogger('PYTATL')
    
class TATLFormula:
    def __init__(self,goal_template,objective):
      """Create a game on the goal template whose objective is:
      - A[phi Until psi] if game_type is true (reachability type: e.g. A[True Until psi]=A<> psi)
      - A[phi WeakUntil psi] if game_type is false (safety type: e.g. A[not phi WeakUntil False]=A[] not phi)
      phi and psi are set of states (dictionaries) in the goal template: each one associate a goal location to a goal zone. 
      """
      self.goal_template = goal_template
      
      if type(objective) == str:
	self.formula = self.parse_formula(objective)
	self.formula_str = objective
      else:
	self.formula = objective.formula
	self.formula_str = objective.formula_str
	
    def parse_formula(self,formula_str):
	#tokens
	
	comma 		= Literal(",")
	lparen 		= Literal("(")
	rparen 		= Literal(")")
	lbrack 		= Literal("[")
	rbrack 		= Literal("]")
	equals 		= Literal("==")
	nequals 	= Literal("!=")
	lessequals 	= Literal("<=")
	less 		= Literal("<")
	greater 	= Literal(">")
	greaterequals	= Literal(">=")
	minus		= Literal("-")
	always		= Literal("A")
	exists		= Literal("E")
	until		= Literal("U")
	weak		= Literal("W")
	orop 		= Keyword("or")
	andop      	= Keyword("and")
	notop       	= Keyword("not")
	trueval       	= Keyword("true")
	falseval       	= Keyword("false")
	fa		= Keyword("Fa")
	fs		= Keyword("Fs")
	fe		= Keyword("Fe")
	inf 		= Keyword("inf")	 
	
	identifier = Word( alphas, alphanums + "_" )
	number = Word(nums)
	integer = Combine( Optional(minus) + number )
	
	# grammar
	
	interval = Group(Suppress(lbrack) + integer + Suppress(comma) + (integer | inf) + Suppress(rbrack))
	
	orExpression = Forward()
	
	feature = Combine((fa|fe|fs) + lbrack + number + rbrack)
	
	atomicExpression = feature | (lparen + orExpression + rparen)
	
	notExpression = (Optional(notop) + atomicExpression)

	andExpression = Forward()
	andExpression << (notExpression + Optional(andop + andExpression))
	
	orExpression << andExpression + Optional(orop + orExpression)
	
	featureFormula = Forward()

	compop = equals | nequals | lessequals | less | greaterequals | greater
	
	guard = Combine((identifier + compop + integer) | (integer + compop + identifier), adjacent=False).setParseAction(lambda t : self.parse_dbm(t[0]))
			
	loc = Combine(identifier).setParseAction(lambda t : self.parse_location(t[0]))
	
	atomicFormula = (trueval.setParseAction(lambda t : self.parse_true_false(True)) | \
			falseval.setParseAction(lambda t : self.parse_true_false(False)) | \
			guard | \
			loc | \
			(Suppress(lparen) + featureFormula + Suppress(rparen))
			)
			
	notFormula = Group(Optional(notop) + atomicFormula)
	
	andFormula = Forward()
	andFormula << Group(notFormula + Optional(andop + andFormula))
	
	orFormula = Forward()
	orFormula << Group(andFormula + Optional(orop + orFormula))
	
	pathFormula = ((Suppress(lparen) + featureFormula + Suppress(rparen) + until + Optional(interval) + Suppress(lparen) + featureFormula + Suppress(rparen)) | \
		       (Suppress(lparen) + featureFormula + Suppress(rparen) + weak + Suppress(lparen) + featureFormula + Suppress(rparen)) | \
		       (less + greater + Optional(interval) + Suppress(lparen) + featureFormula + Suppress(rparen)).setParseAction(lambda t : [self.parse_true_false(True), "U"] + t[2:]) | \
		       (lbrack + rbrack + Suppress(lparen) + featureFormula + Suppress(rparen)).setParseAction(lambda t : [t[2], "W", self.parse_true_false(False)]))
	# | Combine(delta + compop) + integer + featureFormula | \  
	
	stateFormula = (Group(always + pathFormula) | Group(exists + pathFormula) | orFormula)
		
	featureFormula << Group(Optional((Suppress(lbrack) + orExpression + Suppress(rbrack)).setParseAction(lambda t : self.parse_bdd(t))) + stateFormula)
	
	return (featureFormula+stringEnd).parseString(formula_str)
	
	
    #def test(self,string,obj):
      #print string + " " + str(obj) + "\n"
	
    def parse_bdd(self,constraints):
      bdd = self.goal_template.parse_bdd(" ".join(constraints))
      if bdd is None:
	logger.error('Cannot parse boolean expression \'' + " ".join(constraints) + '\' in formula. Aborting.')
        sys.exit(1)
      return bdd
      
    def parse_dbm(self,constraints):
      zone = self.goal_template.parse_dbm(constraints)
      if zone is None:
	logger.error('Cannot parse guard \'' + constraints + '\' in formula. Aborting.')
        sys.exit(1)
	
      phi = {}
      for l in self.goal_template.locations:
	phi[l] =  pynta.SetOfSymbolicStates()
	phi[l].append(pynta.SymbolicState(copy.copy(l.invariant_bdd),zone & l.invariant_dbm))
      return phi
    
    def parse_location(self,loc):
      #print "parse location " + loc + "\n"
      phi = {}
      for l in self.goal_template.locations:
	phi[l] =  pynta.SetOfSymbolicStates()
	if str(l.name) == str(loc):
	  phi[l].append(pynta.SymbolicState(copy.copy(l.invariant_bdd),l.invariant_dbm.copy()))
	else:
	  phi[l].append(pynta.SymbolicState(copy.copy(l.invariant_bdd),self.goal_template.context.getEmptyFederation()))
      return phi

    def parse_true_false(self,val):
      phi = {}
      if val:
	#print "parse true val\n"
	
	for l in self.goal_template.locations:
	  phi[l] =  pynta.SetOfSymbolicStates()
	  phi[l].append(pynta.SymbolicState(copy.copy(l.invariant_bdd),l.invariant_dbm.copy()))
      else:
	#print "parse false val\n"
	
	for l in self.goal_template.locations:
  	  phi[l] =  pynta.SetOfSymbolicStates()
	  phi[l].append(pynta.SymbolicState(copy.copy(l.invariant_bdd),self.goal_template.context.getEmptyFederation()))
      return phi

